/* tif2gds v2

Originally developed around 1995 as the hack hex2cif.c by M. Rooks at Cornell University.
Much later (2011) cleaned up as tif2cif.c and later (2013) becoming tif2gds by M. Rooks 
at Yale University.

The TIFF image file must use 8-bit grayscale.  Bright pixels will be
converted to 10nm boxes, and the boxes will be merged, if possible, into larger
polygons. It's up to you to scale the cell.

The resulting GDSII file can be read directly into Layout or Design Workshop, as well as
many other CAD programs.  

A typical picture of a face should be converted to grayscale (with photoshop or gimp)
then increased in contrast, inverted, and then half-toned (aka "newsprint"), producing
a binary image of dots.

One structure (cell) is created, with the same name as the input file.

v2  Use BYTE instead of BOOL for unsigned pixel data. OMG how did this program
    work at all before? 

*/

#include "libgds.h"

// if you put libgds.h in /usr/include then use the form
// #include <libgds.h>

#include <tiffio.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define VERSION 2

char
    *sPatternName,
    *sOutFileName,
    *sInFileName,
    *sBase;

int
    OutFile; // file descriptor

TIFF
    *InFile;
    
float
    fBit;

int
    iXstart,
    iXend,
    iBoxWidth,
    j,
    iPixelSizeX=1,    
    iPixelSizeY=1,
    iPixelPitchX=1,
    iPixelPitchY=1,
    iXpixel,
    iPrevious,
    iPixel,
    iBit,
    iNx,
    iNy,
    iNhex,
    iX,
    iY,
    iBoxLength,
    iDark,
    i;

char
    by8Pixels;

BOOL
    bBackgroundIsZero=1;


/*--------------------------------------------------------------------------*/


BOOL **create_bool_matrix( nrows, ncols )

    int
        nrows,
        ncols;
    {

    BOOL
        *block,
        **m;

    int i;


    block = (BOOL *) malloc((unsigned) 
            nrows * ncols * sizeof(BOOL*));

    if (! block) 
    printf("\nerror: unable to allocate boolean matrix\n");

    m = (BOOL **) malloc( (unsigned) nrows * sizeof( BOOL* ) );
    
    if (!m) 
        printf( "\nerror: unable to allocate m\n" );
       
    for( i=0; i < nrows; i++ ) 
    m[i] = block + (ncols * i); 


    return m;
    }
/*--------------------------------------------------------------------------*/


BYTE **create_byte_matrix( nrows, ncols )

    int
        nrows,
        ncols;
    {

    BYTE
        *block,
        **m;

    int i;


    block = (BYTE *) malloc((unsigned) 
            nrows * ncols * sizeof(BYTE*));

    if (! block) 
    printf("\nerror: unable to allocate byte matrix\n");

    m = (BYTE **) malloc( (unsigned) nrows * sizeof( BYTE* ) );
    
    if (!m) 
        printf( "\nerror: unable to allocate m\n" );
       
    for( i=0; i < nrows; i++ ) 
    m[i] = block + (ncols * i); 


    return m;
    }

/*-----------------------------------------------------------------------*/

BOOL all_set_between( int iXstart, int iXend, int j, BYTE **baImage ) 

    /* Determine if all the bits are set between the given locations on */
    /* the given row.                                                   */

    {
    int n;

    n = iXstart;
    while ( n < iNx && n <= iXend && baImage[n][j] ) n++;
    return( n > iXend );
    }
 
/*---------------------------------------------------------------------------*/

BOOL not_hit_between( int iXstart, int iXend, int j, BOOL **baHit )

    /* Determine if the pixel map has been "hit" -  that is, if a */
    /* rectangle has already been generated - between the given   */
    /* pixels on the given row.                                   */

    {
    int n;

    n = iXstart;
    while ( n < iNx && n <= iXend && ! baHit[n][j] ) n++;
    return( n > iXend );
    }

/*---------------------------------------------------------------------------*/

void mark_as_hit_between( int iXstart, int iXend, int j, BOOL **baHit )

    /* When we generate a rectangle, we set the bits in the hit matrix */
    /* so that we will not generate any other rectangle over another.  */
    /* This function sets the bits on row j, between the given pixels. */

    {
    int i;

    for ( i=iXstart; i<=iXend; i++ ) baHit[i][j] = TRUE;
    }

/*---------------------------------------------------------------------------*/

BYTE **
read_tiff( char *filename, int *nx, int *ny )
    
    // Lifted from Michael Still's page "Graphics programming with libtiff",
    // https://www.ibm.com/developerworks/linux/library/l-libtiff/
    // with minor adjustments and many thanks.

    {

    TIFF *image;
    uint16 photo, bps, spp, fillorder;
    uint32 width;
    tsize_t stripSize;
    unsigned long imageOffset, result;
    int stripMax, stripCount, r, c;
    char *buffer, *buffer_ptr, tempbyte;
    unsigned long bufferSize, count;
   
    BYTE **image2d, min, max;


    // Open the TIFF image

    if((image = TIFFOpen(filename, "r")) == NULL)
        {
        fprintf(stderr, "\nERROR: could not open image file %s\n", filename );
        exit(42);
        }

    // Check that it is of a type that we support

    if((TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &bps) == 0) || (bps != 8))
        {
        fprintf(stderr, "\nERROR: Unsupported number of bits per sample: %d.", bps );
        fprintf(stderr, "\n       We want grayscale, 8 bit pixels.  \n\n" );
        exit(42);
        }

    if((TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &spp) == 0) || (spp != 1))
        {
        fprintf(stderr, "\nERROR: unsupported number of samples per pixel: %d", spp );
        fprintf(stderr, "\n       We want grayscale, not color.\n\n" );
        exit(42);
        }

    // Read in the possibly multiple strips

    stripSize = TIFFStripSize (image);
    stripMax = TIFFNumberOfStrips (image);

    imageOffset = 0;
  
    bufferSize = TIFFNumberOfStrips (image) * stripSize;

    if((buffer = (char *) malloc(bufferSize)) == NULL)
        {
        fprintf(stderr, "\nERROR: Could not allocate memory for the uncompressed image\n");
        exit(42);
        }
  
    for (stripCount = 0; stripCount < stripMax; stripCount++)
        {
        if((result = TIFFReadEncodedStrip (image, stripCount, buffer + imageOffset, stripSize)) == -1)
            {
            fprintf(stderr, "\nERROR: Read error on input strip number %d\n", stripCount);
            exit(42);
            }

        imageOffset += result;
        }

    // Deal with photometric interpretations

    if(TIFFGetField(image, TIFFTAG_PHOTOMETRIC, &photo) == 0)
        {
        fprintf(stderr, "\nERROR: Image has an undefined photometric interpretation: %d\n", photo );
        exit(42);
        }
  
    if(photo != PHOTOMETRIC_MINISBLACK) // Flip bits
        {
        printf("\nFixing the photometric interpretation...\n");

        for(count = 0; count < bufferSize; count++)
            buffer[count] = ~buffer[count];
        }

    // Deal with fillorder (little endian vs big endian)


    /*........ many TIFF files leave the endian state undefined......

    if(TIFFGetField(image, TIFFTAG_FILLORDER, &fillorder) == 0)
        {
        fprintf(stderr, "\nERROR: Image has an undefined fillorder: %d\n", fillorder );
        exit(42);
        }
  
    if(fillorder != FILLORDER_MSB2LSB) // We need to swap bits -- ABCDEFGH becomes HGFEDCBA
        {
        printf("\nFixing the fillorder...\n");

        for(count = 0; count < bufferSize; count++)
            {
            tempbyte = 0;
            if( buffer[count] & 128 ) tempbyte += 1;
            if( buffer[count] & 64  ) tempbyte += 2;
            if( buffer[count] & 32  ) tempbyte += 4;
            if( buffer[count] & 16  ) tempbyte += 8;
            if( buffer[count] & 8   ) tempbyte += 16;
            if( buffer[count] & 4   ) tempbyte += 32;
            if( buffer[count] & 2   ) tempbyte += 64;
            if( buffer[count] & 1   ) tempbyte += 128;
            buffer[count] = tempbyte;
            }
        }
     ................................... and so this must be disabled */
 

    if(TIFFGetField(image, TIFFTAG_IMAGEWIDTH, &width) == 0)
        {
        fprintf(stderr, "\nERROR: Image does not define its width\n");
        exit(42);
        }


    // Do whatever voodo you do  .....................................


    TIFFGetField( image, TIFFTAG_IMAGEWIDTH,  nx  );
    TIFFGetField( image, TIFFTAG_IMAGELENGTH, ny  );

    printf( "\nNumber of rows:    %d",   *ny );
    printf( "\nNumber of columns: %d\n", *nx );

    if ( (*nx) * (*ny) > bufferSize )
        {
        printf( "\nERROR: number of pixels is greater than the buffer size\n" );
        exit(0);
        }

    image2d = create_byte_matrix( *nx, *ny );

    buffer_ptr = buffer;

    min = *buffer_ptr;
    max = min;

    for ( r=(*ny-1); r>=0; r-- )
        {
        for ( c=0; c < *nx; c++ )
            {
            image2d[c][r] = *buffer_ptr;
            buffer_ptr++;
            if ( image2d[c][r] > max ) max = image2d[c][r];
            if ( image2d[c][r] < min ) min = image2d[c][r];
            }
        }

    TIFFClose(image);

    //printf( "\nmin = %d   max = %d\n", min, max );

    return( image2d );

    }

/*---------------------------------------------------------------------------*/


main( int argc, char *argv[] )
    {

    struct coord
        {
        int x;
        int y;
        };

    struct coord
        p1,
        p2,
        p3,
        p4;

    int
        i,
        iRowsPerStrip,
        iMin,
        iMax,
        scale,
        x[5],
        y[5];

    char
        ch;

    BYTE
      **baImage;

    BOOL
      **baHit;


    sBase        = malloc( 296 * sizeof( char ) );
    sInFileName  = malloc( 300 * sizeof( char ) );
    sOutFileName = malloc( 300 * sizeof( char ) );
    sPatternName = malloc( 300 * sizeof( char ) );

    if( argc > 1 )
        {
        if( strcmp( argv[1], "-h" ) == 0 || strcmp( argv[1], "--help" ) == 0 )
            {
            printf( "\nTIFF to GDSII converter, by M. Rooks, Yale University.\n" );
            printf( "Version %d\n", VERSION );
            printf( "\nUse: tif2gds file\n" );
            printf( "\n     where the file name 'file' does not include the suffix '.tif'" );
            printf( "\n     The output file will be file.gds\n\n" );
            exit(0);
            }
        sBase = argv[1];
        }
    else
        {
        printf( "\nEnter input file name, without the .tif suffix > " );
        scanf( "%s", sBase );
        }    


    strcpy( sInFileName, sBase );
    strcat( sInFileName, ".tif" );
    printf(                       "Input file:           %s\n", sInFileName );
    
    strcpy( sOutFileName, sBase );
    strcat( sOutFileName, ".gds" );
    printf(                       "Output file:          %s\n", sOutFileName );

    OutFile = open( sOutFileName, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );

    if ( OutFile <= 0 )
        {
        printf( "\nERROR: unable to open output file.\n" );
        exit( -1 );
        }


    baImage = read_tiff( sInFileName, &iNx, &iNy );

    baHit   = create_bool_matrix( iNx, iNy );

    iMax = baImage[0][0];
    iMin = iMax;

    for ( iY=0; iY < iNy; iY++ )
        {
        for ( iX=0; iX < iNx; iX++ )
            {
            baHit[iX][iY] = FALSE;
            if ( baImage[iX][iY] > iMax ) iMax = baImage[iX][iY];
            if ( baImage[iX][iY] < iMin ) iMin = baImage[iX][iY];
            }     
        }      

    if ( iMin < 0 || iMax < 0 || iMax == iMin )
        {
        printf( "\nERROR: strange or blank image.\n" );
        exit(0);
        }

    iDark = (iMax + iMin) / 4;

    // printf( "\nThreshold: %x\n", iDark );

    for ( iY=0; iY < iNy; iY++ )
        {
        for ( iX=0; iX < iNx; iX++ )
            {
            if ( baImage[iX][iY] < iDark ) 
	        baImage[iX][iY] = 0;
            else
                baImage[iX][iY] = 1;
            }     
        }         


    // Write the GDSII file header
    // Create cell with the name in sBase 

    gds_create_lib( OutFile, sBase, 0.001 /* um per bit */ );
    gds_write_bgnstr( OutFile );
    gds_write_strname( OutFile, sBase );

    printf( "\n" );

    /*''''''''''''''''''''''''''''''''''''''''''''''''*/
    /* now search through the image for blocks of 1's */
    /*,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,*/

    for ( iY=0; iY < iNy; iY++ )
        {
        if ( ! (iY % 50) )
            {
            printf( "\015Converting row %1d of %d            ", iY+1, iNy );
            fflush( stdout );
            }

        i = 0;
        while ( i < iNx )
            {
            iBoxLength = 0;
            iXstart = i;
            iXend   = i;
            while ( i < iNx && baImage[i][iY] && ! baHit[i][iY] )
                {
                iXend = i;
                iBoxLength += iPixelSizeX;
                i++;
                }
            if ( iBoxLength > 0 ) /* see how deep it is */
                {
                iBoxWidth = 0;
                j = iY;

                while ( j < iNy &&
                        all_set_between( iXstart, iXend, j, baImage ) && 
                        not_hit_between( iXstart, iXend, j, baHit   )    )
                    {
                    iBoxWidth += iPixelSizeY;
                    mark_as_hit_between( iXstart, iXend, j, baHit );
                    j++;
                    }

                scale = 1000;

                p1.x = iXstart * scale;
                p1.y = iY      * scale;
                p2.x = p1.x;
                p2.y = p1.y + iBoxWidth  * scale;
                p3.x = p1.x + iBoxLength * scale;
                p3.y = p2.y;
                p4.x = p3.x;
                p4.y = p1.y;

                gds_write_boundary( OutFile );       // write the token
                gds_write_layer( OutFile, 0 );       // layer 0
                gds_write_datatype( OutFile, 0 );    // datatype 0

                x[0] = p1.x;  y[0] = p1.y;     
                x[1] = p2.x;  y[1] = p2.y;       
                x[2] = p3.x;  y[2] = p3.y;     
                x[3] = p4.x;  y[3] = p4.y;
                x[4] = p1.x;  y[4] = p1.y;     

                gds_write_xy( OutFile, x, y, 5 ); 
                gds_write_endel( OutFile );   

                }
            i++;
    
            }      /* end while  */
        }          /* end for iY */
        
    gds_write_endstr( OutFile );
    gds_write_endlib( OutFile );

    close( OutFile );

    printf( "\n\nDone.\n" );

    }






